use std::fs;
use std::io::ErrorKind;
use std::os::unix::net::{UnixListener, UnixStream};
use std::path::{Path, PathBuf};
use std::thread;

use color_eyre::eyre::Context;
use color_eyre::Result;
use crossbeam::channel::{unbounded, Receiver};

type Task = Result<UnixStream>;

pub struct Listener {
    path: PathBuf,
    receiver: Receiver<Task>,
}

impl Listener {
    pub fn new(runtime_dir: impl AsRef<Path>, session: String) -> Result<Self> {
        let runtime_dir = runtime_dir.as_ref();

        if let Err(err) = fs::metadata(runtime_dir) {
            if err.kind() == ErrorKind::NotFound {
                fs::create_dir_all(runtime_dir).wrap_err_with(|| {
                    format!(
                        "Failed to create runtime directory {}",
                        runtime_dir.display()
                    )
                })?;
            }
        }

        let path = runtime_dir.join(session).with_extension("s");
        let listener = UnixListener::bind(&path)
            .wrap_err_with(|| format!("Failed to bind listener to {}", path.display()))?;

        tracing::info!("Listening on {}", path.display());

        let (tx, rx) = unbounded();

        thread::Builder::new()
            .name("socker-listener".into())
            .spawn(move || {
                for stream in listener.incoming() {
                    tx.send(stream.map_err(From::from))
                        .expect("Failed to send stream");
                }
            })
            .wrap_err("Failed to spawn listener thread")?;

        Ok(Self { path, receiver: rx })
    }

    pub fn receiver(&self) -> &Receiver<Task> {
        &self.receiver
    }

    pub fn close(&mut self) -> Result<()> {
        tracing::trace!("Removing socket '{}'", self.path.display());

        fs::remove_file(&self.path)
            .wrap_err_with(|| format!("Failed to remove socket file '{}'", self.path.display()))
    }
}

impl Drop for Listener {
    fn drop(&mut self) {
        self.close().expect("Failed to close Listener")
    }
}
