use std::io::Read;
use std::os::unix::net::UnixStream;
use std::sync::atomic::{AtomicBool, Ordering};
use std::sync::Arc;
use std::thread::{self, JoinHandle};
use std::time::Duration;
use std::{fs, iter};

use color_eyre::eyre::{self, Context};
use color_eyre::Result;
use crossbeam::deque::{Injector, Stealer, Worker};

use crate::kakoune::editor_quote;
use crate::Request;

type Task = Result<UnixStream>;

pub struct TaskScheduler {
    injector: Arc<Injector<Task>>,
    terminate: Arc<AtomicBool>,
    threads: Vec<JoinHandle<()>>,
}

impl TaskScheduler {
    pub fn new(workers: u8) -> Result<Self> {
        let terminate = Arc::new(AtomicBool::new(false));

        let injector = Arc::new(Injector::new());
        let workers: Vec<_> = (0..workers).map(|_| Worker::new_fifo()).collect();
        let stealers: Vec<_> = workers.iter().map(|w| w.stealer()).collect();
        let stealers = Arc::new(stealers);

        let threads = workers
            .into_iter()
            .enumerate()
            .map(|(i, worker)| {
                let injector = injector.clone();
                let stealers = stealers.clone();
                let terminate = terminate.clone();

                thread::Builder::new()
                    .name(format!("worker-{}", i))
                    .spawn(|| thread_handler(worker, injector, stealers, terminate))
                    .map_err(From::from)
            })
            .collect::<Result<Vec<_>>>()
            .wrap_err("Failed to spawn worker threads")?;

        tracing::info!("Started {} worker threads", threads.len());

        Ok(Self {
            injector,
            terminate,
            threads,
        })
    }

    pub fn schedule(&mut self, task: Task) {
        self.injector.push(task)
    }

    pub fn terminate(self) {
        self.terminate.store(true, Ordering::Relaxed);

        for handle in self.threads {
            if let Err(err) = handle.join() {
                tracing::error!("Worker thread panicked: {:?}", err);
            }
        }
    }
}

#[tracing::instrument]
fn find_task<T>(local: &Worker<T>, global: &Injector<T>, stealers: &[Stealer<T>]) -> Option<T> {
    local.pop().or_else(|| {
        iter::repeat_with(|| {
            global
                .steal_batch_and_pop(local)
                .or_else(|| stealers.iter().map(|s| s.steal()).collect())
        })
        .find(|s| !s.is_retry())
        .and_then(|s| s.success())
    })
}

#[tracing::instrument]
fn thread_handler(
    worker: Worker<Task>,
    injector: Arc<Injector<Task>>,
    stealers: Arc<Vec<Stealer<Task>>>,
    terminate: Arc<AtomicBool>,
) {
    loop {
        let task = 'find_task: loop {
            if terminate.load(Ordering::Relaxed) {
                return;
            }

            if let Some(task) = find_task(&worker, &injector, &stealers) {
                break 'find_task task;
            }

            thread::sleep(Duration::from_millis(50));
        };

        if let Err(err) = handle_connection(task) {
            tracing::error!("{:?}", err);
        }
    }
}

#[tracing::instrument(skip_all)]
fn handle_connection(task: Task) -> Result<()> {
    let mut stream = task.wrap_err("Failed to receive client connection")?;

    let mut buf = String::new();
    stream
        .read_to_string(&mut buf)
        .wrap_err("Failed to read from connection")?;
    let req: Request = toml::from_str(&buf).wrap_err("Failed to parse request")?;

    tracing::info!("Received request: {:?}", req);

    let response = process_request(&req)
        .unwrap_or_else(|err| format!("fail {}", editor_quote(format!("{}", err))));

    tracing::debug!("Sending response:\n{}", response);

    fs::write(&req.fifo, response.as_bytes()).wrap_err("Failed to write to command fifo")?;

    Ok(())
}

#[tracing::instrument]
fn process_request(req: &Request) -> Result<String> {
    eyre::bail!("Not implemented")
}
