use std::fs::File;
use std::io::stderr;
use std::path::PathBuf;

use clap::{Parser, Subcommand};
use color_eyre::eyre::Context;
use color_eyre::Result;
use serde::Deserialize;
use tracing::metadata::LevelFilter;
use tracing_error::ErrorLayer;
use tracing_subscriber::fmt;
use tracing_subscriber::prelude::*;

mod client;
mod daemon;
mod kakoune;

#[derive(Parser)]
#[command(author, version, about, long_about = None)]
struct Cli {
    /// Set the level of verbosity
    #[arg(short, long, action = clap::ArgAction::Count)]
    verbose: u8,
    /// File to write log events to
    #[arg(long, value_name = "FILENAME")]
    log: Option<PathBuf>,
    /// The command to run
    #[command(subcommand)]
    command: Command,
}

#[derive(Subcommand)]
enum Command {
    /// Send a request payload to a running daemon and wait for the response
    Request,
    /// Start a daemon
    Daemon {
        /// The Kakoune session this daemon belongs to
        session: String,
        /// The number of worker threads to spawn
        #[arg(short = 'n', long, default_value = "2", action = clap::ArgAction::Count)]
        workers: u8,
    },
}

#[derive(Clone, Debug, Deserialize)]
struct Request {
    /// The Kakoune session this request came from
    session: String,
    /// The command FIFO provided by Kakoune
    fifo: String,
}

#[tracing::instrument]
fn main() -> Result<()> {
    let cli = Cli::parse();

    {
        let filter_layer = match cli.verbose {
            0 => LevelFilter::WARN,
            1 => LevelFilter::INFO,
            2 => LevelFilter::DEBUG,
            _ => LevelFilter::TRACE,
        };

        let stderr_layer = if cfg!(debug_assertions) {
            fmt::layer().pretty().with_writer(stderr).boxed()
        } else {
            fmt::layer().compact().with_writer(stderr).boxed()
        };
        let file_layer = if let Some(path) = cli.log {
            let f = File::create(&path)
                .wrap_err_with(|| format!("Failed to create log file '{}'", path.display()))?;
            let layer = fmt::layer().pretty().with_writer(f);
            Some(layer)
        } else {
            None
        };

        tracing_subscriber::registry()
            .with(filter_layer)
            .with(stderr_layer)
            .with(file_layer)
            .with(ErrorLayer::new(fmt::format::Pretty::default()))
            .init();
    }

    let runtime_dir = std::env::var("XDG_RUNTIME_DIR")
        .map(PathBuf::from)
        .unwrap_or_else(|_| "/var".into())
        .join("kak-highlight");

    match cli.command {
        Command::Daemon { session, workers } => daemon::handle(runtime_dir, session, workers),
        Command::Request => client::handle(runtime_dir),
    }
}
